/*
 * mping -- parallel ping
 */

#include <stdio.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <netdb.h>
#include <string.h>

/* Design
 *
 * This reads a configuration file (or stdin) and gets a list of nodes
 * to ping, together with options for packet size and retry count.
 *
 * It pings all these in parallel until the entire list is satisfied
 * or all retry counts are exhausted.
 *
 * We do this by setting an interval timer for the basic inter-packet
 * interval and keeping a process clock.  Every time a system is ready
 * to be sent a packet, we wait for the next process clock interval
 * (to avoid overloading the network).
 */

struct plist {
	struct plist*	next;
	char		name[32];
	struct sockaddr_in target;
	unsigned long	nextSend;	/* virtual clock time for next send */
	int		timeOut;	/* inter-packet transmit interval */
	int		numRetries;	/* # of retries to make */
	int		numTried;	/* # sent on this ping */
	int		numToSend;	/* # of successful returns req'd */
	int		numSent;	/* # of packets sent */
	int		numReceived;	/* # of responses received */
	int		dataPattern;	/* what to fill packet with */
	int		status;
#define STS_WAITING	0x0		/* waiting for response or next turn */
#define STS_DONE	0x1		/* success or timeout */
#define STS_READY	0x2		/* waiting to ping */
#define STS_ACTIVE	0x3		/* pinging */
	int		pktSize;
	char*		packet;
	unsigned long	pktTimes;	/* total microseconds of packet RTT */ };

/* This defines, in milliseconds, the basic inter-packet interval we
 * want to maintain for the parallel pings.  This ignores round-off
 * problems with the system clock.
 */
#define TICK	50			/* milliseconds per TICK */
#define INV_TICK	(1000/TICK)	/* TICKs per second */
#define RETRY_TIME	(1000/TICK)	/* 1 second retry interval */
char*	createPacket();

/* Defaults for optional arguments */
int	dflt_size	=   56;	/* 56 bytes of ICMP packet data */
int	dflt_retries	=    5;	/* make 5 attempts before quitting */
int	dflt_toSend	=    1;	/* quit on 1 successful return */
int	dflt_pattern	=    0;	/* packet data pattern */
int	dflt_timeOut	= 1000;	/* in milliseconds */

/* Minima for optional arguments */
int	min_data_size	= sizeof (struct timeval) + sizeof(short);
int	min_timeOut	= 100;	/* milliseconds */

struct plist*	sendq;		/* targets ready to send */
struct plist*	plist_array;	/* all configured targets */
int	num_plist;		/* count of configured targets */
int	siz_plist;		/* count of allocated plist structs */
int	num_active;		/* number of targets still active */

/* The socket handle */
int	pingSocket;
int	verbose;
int	icmp_ident;

#define HUNK	64
#define FOREACHPLIST(P,I) \
	for((P)=plist_array,(I)=0; (I)<num_plist; (I)++,(P)++)

extern int	errno;
extern int	sys_nerr;
extern char*	sys_errlist[];
#define STRERROR	(errno>=sys_nerr ? "unknown" : sys_errlist[errno])

/* return pointer to newly-allocated plist */
static struct plist*
new_plist() {
	unsigned nbytes;
	unsigned oldsize;
	if (num_plist >= siz_plist) {
		oldsize = siz_plist;
		siz_plist += HUNK;
		nbytes = siz_plist * sizeof (struct plist);
		plist_array = (struct plist*) (plist_array
			? realloc (plist_array, nbytes)
			: malloc (nbytes));

		/* clear the new space */
		bzero ((char*)&plist_array[oldsize], HUNK*sizeof(struct plist));
	}
	return &plist_array[num_plist++];
}

add_plist_byTime (head, item)
struct plist** head;
struct plist* item;
{
	register struct plist** q;
	register struct plist* p;

	for (q = head; p = *q; q = & (*q)->next)
		if (p->nextSend > item->nextSend)
			break;

	item->next = p;
	*q = item;
}

struct plist*
get_plist (q)
struct plist**q;
{
	struct plist* p;
	if ((p = *q) != 0) {
		*q = p->next;
		p->next = 0;
	}
	return p;
}
	

readConfig (stream)
FILE*	stream;
{
	char buf[512];
	char *vec[10];
	char *ovec[6];
	struct plist* p;
	int nf, no;

	/* name address size retries */
	while (fgets(buf, sizeof buf, stream)) {
		if (buf[0] == '#' || buf[0]== '\n')
			continue;
		nf = strsplit (buf, vec, 10, 0);
		if (nf < 2)
			continue;
		p = new_plist();
		strncpy (p->name, vec[0], sizeof p->name);

		p->pktSize = dflt_size;
		p->numRetries = dflt_retries;
		p->dataPattern = dflt_pattern;
		p->numToSend = dflt_toSend;
		p->timeOut = dflt_timeOut/TICK;
		p->numTried = 0;
		if (nf >= 3) {
			bzero (ovec, sizeof ovec);
			no = strsplit (vec[2], ovec, 6, ",");
			if (ovec[0]) p->pktSize = atoi(ovec[0]);
			if (p->pktSize < min_data_size)
				p->pktSize = min_data_size;

			if (ovec[1]) p->dataPattern = strtol (ovec[1],0,16);
			if (ovec[2]) p->numToSend = atoi (ovec[2]);
			if (p->numToSend <= 0)
				p->numToSend = dflt_toSend;

			if (ovec[3]) p->numRetries = atoi (ovec[3]);
			if (p->numRetries <= 0)
				p->numRetries = dflt_retries;

			if (ovec[4]) p->timeOut = atoi (ovec[4])/TICK;
			if (p->timeOut <= min_timeOut/TICK)
				p->timeOut = min_timeOut/TICK;
		}
		p->packet = (char*)createPacket (p->pktSize, p->dataPattern);
		p->target.sin_family = AF_INET;

		/* If the 2nd argument isn't an address, look it up.
		 * clone the entry for each address.
		 */
		p->target.sin_addr.s_addr = inet_addr (vec[1]);
		if (p->target.sin_addr.s_addr == 0xffffffff) {
			struct hostent *hp;
			struct plist* p2;
			char** addrp;

			if ((hp = gethostbyname (vec[1])) == 0)
				continue;

			addrp = hp->h_addr_list;
			p2 = p;
			while (*addrp) {
				if (p2 == 0) {
					p2 = new_plist();
					*p2 = *p;
				}
				bcopy (*addrp, (caddr_t)&p->target.sin_addr,
					hp->h_length);
				p2 = 0;
				addrp++;
			}
		}
	}
}


/* Dump out the config file for debugging */
void
printConfig ()
{
	register struct plist* p;
	register i;

	FOREACHPLIST(p,i) {
		printf ("plist[%d]: addr=%s; retry=%d; size=%d\n",
			i, inet_ntoa(p->target.sin_addr),
			p->numRetries, p->pktSize);
	}
}




struct timeval clock_interval = { 0, TICK*1000 };

static unsigned long clock_value;
static void clockService()
{
	clock_value++;
}

startTimer()
{
	struct itimerval tim;
	signal (SIGALRM, clockService);

	tim.it_interval = clock_interval;
	tim.it_value = clock_interval;
	setitimer (ITIMER_REAL, &tim, (int*)0);
}

stopTimer()
{
	struct timeval zero;
	struct itimerval tim;

	zero.tv_sec = 0;
	zero.tv_usec = 0;
	tim.it_interval = zero;
	tim.it_value = zero;
	setitimer (ITIMER_REAL, &tim, (int*)0);
}

main(argc, argv)
int	argc;
char**	argv;
{
	FILE* configf = stdin;
	struct plist* target;

#define EQ(S)	(strcmp((S),*argv)==0)
	for (argc--,argv++; argc > 0; argc--, argv++) {
		if (EQ("-f")) {
			configf = fopen(argv[1], "r");
		}
		if (EQ("-v")) {
			verbose++;
		}
	}
#undef EQ

	/* Read configuration.  Allocate plist structures and everything
	 * else necessary to ping all the targets.
	 *
	 * Cycle through the list pinging each one until satisfied or
	 * retry limit exceeded.
	 *
	 * Print results to standard output.
	 */
	readConfig (configf);
	pingSocket = initSocket();
	startTimer();
	readPause();
	while (num_active > 0) {
		if (sendq == 0) {
			readPause();
			continue;
		}
		target = get_plist(&sendq);
		target->status = STS_ACTIVE;
		sendPing (target, target-plist_array);
	}
	stopTimer();

	printStatus();
}

sendPing (target, index)
struct plist* target;
int	index;
{
	static lastSent;
	int pktlen;
	int i;

	if (target->numTried >= target->numRetries
	    || target->numReceived >= target->numToSend) {
		target->status = STS_DONE;
		return 0;
	}

	/* make sure at least 1 tick has elapsed since last packet
	 * transmitted.
	 */
	while (clock_value <= lastSent)
		readPause();

	target->numSent++;
	target->numTried++;
	target->nextSend = clock_value + target->timeOut;
	lastSent = clock_value;
	
	initPacket (target, index);
	if (verbose) {
		printf ("ping %-15s %d (%d/%d of %d/%d)\n",
			inet_ntoa(target->target.sin_addr),
			target->pktSize,
			target->numTried, target->numRetries,
			target->numSent, target->numReceived);
	}
	pktlen = target->pktSize + ICMP_MINLEN;
	i = sendto (pingSocket, target->packet, pktlen, 0,
			&target->target, sizeof (struct sockaddr));
	if (i < 0) {
		fprintf (stderr, "sendto: %s\n", STRERROR);
		target->status = STS_DONE;
	}
	else
		target->status = STS_WAITING;
}

readPause()
{
	register struct plist* p;
	register int i;
	unsigned long old_clock = clock_value;
	struct sockaddr_in from;
	int fromlen ;
	char pkt[4096];
	int len = sizeof pkt;
	fd_set fdmask;

	/* This will block if nothing there */
	for (;;) {
		struct timeval timeout;
		bzero (&fdmask, sizeof fdmask);
		FD_SET(pingSocket, &fdmask);
		timeout.tv_sec = 0; timeout.tv_usec = TICK*1000;
		i = select (pingSocket+1,&fdmask,(int*)0,(int*)0,&timeout);
		if (i > 0 && FD_ISSET(pingSocket, &fdmask)) {
			fromlen = sizeof from;
			i = recvfrom (pingSocket, pkt, len, 0, &from, &fromlen);
			if (i < 0) {
				fprintf ("recvfrom: %s\n", STRERROR);
				break;
			}
			pr_pack (pkt, &from);
		}
		if (clock_value > old_clock)
			break;
	}

	num_active = 0;
	FOREACHPLIST(p,i) {
		if (p->status == STS_DONE)
			continue;
		num_active++;
		if (p->status == STS_WAITING && p->nextSend <= clock_value) {
			p->status = STS_READY;
			add_plist_byTime (&sendq, p);
		}
	}
}

printStatus()
{
	register struct plist*p;
	register i;
	unsigned long rtt;

	FOREACHPLIST (p,i) {
		rtt = p->numReceived ? p->pktTimes/p->numReceived : 0;
		printf ("%-24s %-15s %2d %2d %8lu\n", p->name,
			inet_ntoa(p->target.sin_addr), p->numSent,
			p->numReceived, rtt);
	}
}

/* IP specific routines */


/* The "size" argument here references the ICMP_DATA portion of the
 * packet.  We add to that the ICMP overhead.
 */
char* createPacket (size)
int size;
{
	return malloc (size + ICMP_MINLEN);
}

int initPacket (packet, i)
struct plist* packet;
int	i;
{
	register struct icmp *icp = (struct icmp*) packet->packet;
	u_short *usp;
	struct timeval* tvp;
	struct timezone tz;

	icp->icmp_type = ICMP_ECHO;	
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_seq = packet->numSent;

	if (icmp_ident == 0)
		icmp_ident = getpid() & 0xffff;
	icp->icmp_id = icmp_ident;

	/* fill packet with argument pattern */
	memset (icp->icmp_data, packet->dataPattern, packet->pktSize);

	tvp = (struct timeval*) icp->icmp_data;
	gettimeofday (tvp, &tz);

	/* store the index of this plist structure in the packet for
	 * fast lookup when it comes back.
	 */
	usp = (u_short*) (icp->icmp_data + sizeof (struct timeval));
	*usp = i;

	icp->icmp_cksum = in_cksum ((u_short*)icp, packet->pktSize+ICMP_MINLEN);
}

/*
 *			I N _ C K S U M
 *
 * Checksum routine for Internet Protocol family headers (C Version)
 *
 */
in_cksum(addr, len)
u_short *addr;
int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while( nleft > 1 )  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if( nleft == 1 ) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}

/*
 * 			T V S U B
 * 
 * Subtract 2 timeval structs:  out = out - in.
 * 
 * Out is assumed to be >= in.
 */
tvsub( out, in )
register struct timeval *out, *in;
{
	if( (out->tv_usec -= in->tv_usec) < 0 )   {
		out->tv_sec--;
		out->tv_usec += 1000000;
	}
	out->tv_sec -= in->tv_sec;
}

int
initSocket ()
{
	int s;
	struct protoent* proto;

	proto = (struct protoent*)getprotobyname("icmp");
	s = socket (AF_INET, SOCK_RAW, proto->p_proto);
	if (s < 0) {
		fprintf (stderr, "Cannot create socket: %s\n", STRERROR);
		exit (1);
	}
	return s;
}

pr_pack (pkt, fromptr)
char*	pkt;
struct sockaddr_in *fromptr;
{
	struct icmp *icp;
	struct ip * ipp;
	register struct plist* p;
	register int i;
	struct timeval tv, *tvp;
	struct timezone tz;
	unsigned short *usp;
	unsigned long rtt;

	ipp = (struct ip*) pkt;
	icp = (struct icmp*) (pkt + (ipp->ip_hl * 4));

	/* discard stuff that's not ours */
	if (icp->icmp_type != ICMP_ECHOREPLY || icp->icmp_id != icmp_ident) {
		if (verbose)
			printf ("%s -- %d,%d; not mine\n",
				inet_ntoa(fromptr->sin_addr),
						icp->icmp_type, icp->icmp_id);
		return;
	}

	gettimeofday (&tv, &tz);
	tvp = (struct timeval*)icp->icmp_data;
	/* need to calculate RTT */
	tvsub (&tv, tvp);
	rtt = (tv.tv_sec * 1000000) + tv.tv_usec;

	/* See if it contains a valid index */
	usp = (u_short*)(icp->icmp_data + sizeof (struct timeval));
	i = *usp;
	if (i >= 0 && i < num_plist) {
		p = &plist_array[i];
		if (p->target.sin_addr.s_addr == fromptr->sin_addr.s_addr) {
	 gotit:
			p->numReceived++;
			p->numTried = 0;
			p->pktTimes += rtt;
			if (p->numReceived >= p->numToSend)
				p->status = STS_DONE;
			return;
		}
	}

	/* otherwise, walk through the list of eligibles looking
	 * for a match.
	 */
	FOREACHPLIST(p,i) {
		if (p->target.sin_addr.s_addr == fromptr->sin_addr.s_addr) {
			goto gotit;
		}
	}
	if (verbose) 
		printf ("%s -- %d; not in list\n",
			inet_ntoa(fromptr->sin_addr), icp->icmp_type);
}

/* A version of my strsplit function which uses the new
 * ANSI standard function 'strtok'.  This one doesn't
 * have the behavior of concatenating all remaining tokens
 * in the the last return when <nvec> is less than the
 * number of real <sep>-delimited fields.
 */
int
strsplit (buf, vec, nvec, sep)
char*	buf;
char**	vec;
int	nvec;
char*	sep;
{
	int n = 0;

	if (sep == NULL) sep = " \t\r\n\f";

	while (n < nvec) {
		if ((vec[n] = strtok (buf, sep)) == NULL)
			break;
		if (n == 0) buf = NULL;
		n++;
	}

	/* NULL-terminate pointer array if space permits */
	if (n < nvec)
		vec[n] = NULL;
	return n;
}
